/*
   Direct3D 9 Shader Demo.
   Created by the Programming Ace.
   www.UltimateGameProgramming.com

   The purpose of this demo is to create a basic vertex shader in Direct3D.
   This demo will translate the triangle through the shader as well as set
   each vertex colors in the shader.  Read through VertexShader1.vsh (the shader)
   to see the vertex shader code.  If you are new to vertex shaders then I
   recommend that you read my Vertex/Pixel Shader Introduction article that is posted
   on the web site.  It will provide you with all you need to know about how to write
   a vertex shader in Direct3D/OpenGL.  This demo is meant to be an exercise after
   reading the presentation.

   In order to run this demo you will need DirectX 9.0.  You can obtain this from
   Microsoft's website.  Next you will need a video board that supports vertex/pixel
   shaders.  If you board does support it but the demo does not work, you might
   want to look at upgrading your graphics drivers (OpenGL users).  These drivers
   can be downloaded from your video card marker's website.  If you have a video card
   that supports it and DirectX 9.0 (if you are using the OpenGL demo then you will
   need OpenGL 1.4) on your PC then you are ready to go.
*/


// Demo: Direct3D 9 Shader.


#define WIN32_LEAN_AND_MEAN
#define VC_LEANMEAN

#include<d3dx9.h>


// Our vertex stucture define.
#define D3DFVF_D3DVertex (D3DFVF_XYZ | D3DFVF_DIFFUSE)


// Function Prototypes...
bool InitializeDirect3D(HWND hwnd, bool fullscreen);
bool InitializeObject();
void RenderScence();
void ShutdownDirect3D();


// Direct3D objects.
LPDIRECT3D9 Direct3D_Object = NULL;
LPDIRECT3DDEVICE9 D3D_Device = NULL;
LPDIRECT3DVERTEXBUFFER9 Vertex_Buffer = NULL;


// These are our vertex shader and vertex declaration.
LPDIRECT3DVERTEXSHADER9 VertexShader = NULL;
LPDIRECT3DVERTEXDECLARATION9 ShaderDeclaration = NULL;


// Our 3 matrices...
D3DXMATRIX WorldMatrix;
D3DXMATRIX ViewMatrix;
D3DXMATRIX ProjectionMatrix;


// Vertex stucture.
struct Vertex
{
   FLOAT x, y, z;
   DWORD color;
};


// Shader declaration.
D3DVERTEXELEMENT9 SDec[] =
{
   {0, 0, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0},
	{0, 12, D3DDECLTYPE_D3DCOLOR, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_COLOR, 0},
   D3DDECL_END()
};


LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch(message)
       {
		   case WM_KEYUP:
            // If the user presses the escape key then exit the application.
            if(wParam == VK_ESCAPE)
               PostQuitMessage(0);
            break;

         case WM_CLOSE:
         case WM_DESTROY:
            PostQuitMessage(0);
            break;

            default:
               break;
       }

    return DefWindowProc( hwnd, message, wParam, lParam );
}


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
   MSG msg;                                           // Message object.
   HWND hwnd;                                         // Handle to the window.
   WNDCLASSEX windowClass;                            // Window class object.
   bool done = false;                                 // True then exit.

   // This is the Window class.
   windowClass.cbSize = sizeof(WNDCLASSEX);           // size of the WNDCLASSEX structure.
   windowClass.style = CS_HREDRAW | CS_VREDRAW;       // style of the window.
   windowClass.lpfnWndProc = WndProc;                 // Address to the windows procedure.
   windowClass.cbClsExtra = 0;                        // Extra class information.
   windowClass.cbWndExtra = 0;                        // Extra window information.
   windowClass.hInstance = hInstance;                 // Handle of application Instance.
   windowClass.hIcon = LoadIcon(NULL, IDI_APPLICATION);// Window Icon.
   windowClass.hCursor = LoadCursor(NULL, IDC_ARROW); // Load mouse cursor.
   windowClass.hbrBackground = NULL;                  // Background color.
   windowClass.lpszMenuName = NULL;                   // Menu.
   windowClass.lpszClassName = "UGPClass";            // Name of the window class.
   windowClass.hIconSm = LoadIcon(NULL, IDI_APPLICATION);// Minimized window icon.

   // You must register you class with Windows.
   if(!RegisterClassEx(&windowClass)) return 0;

   // Create the window.
   hwnd = CreateWindowEx(NULL,                                       // The extended style.
                         "UGPClass",                                 // Window class.
                         "Direct3D Vertex Shader 1 Demo by The Programming Ace",// Window name.
                          WS_OVERLAPPEDWINDOW | WS_VISIBLE |         // Window style.
                          WS_SYSMENU |WS_CLIPCHILDREN |              // Window style.
                          WS_CLIPSIBLINGS,                           // Window style.
                          100, 100,                                  // X, Y coords.
                          640, 480,                                  // Window size.
                          NULL,                                      // Handle to parent window.
                          NULL,                                      // Menu.
                          hInstance,                                 // Handle to app instance.
                          NULL);                                     // Pointer to window.

   // If there was an error with creating the window, then close the program.
   if(!hwnd) return 0;


   ShowWindow(hwnd, SW_SHOW);    // Show the window.
   UpdateWindow(hwnd);           // Update its display.

   done = false;                 // false = run program, true means stop.

   // Initialize Direct3D.  If fails then we don't want to run the application.
   if(!InitializeDirect3D(hwnd, false)) done = true;
   
   // Application loop.
   while(!done)
      {
         if(PeekMessage(&msg, NULL, NULL, NULL, PM_REMOVE))
            {
               // If a quit message is received then stop rendering and quit the app.
               if(msg.message == WM_QUIT)
                  {
                     done = true;
                  }

               TranslateMessage(&msg);  // Translate any messages.
               DispatchMessage(&msg);   // Dispatch any messages.
            }
         else
            {
               RenderScence();          // Render a frame.
            }
      }


    // Here we shutdown Direct 3D.
	ShutdownDirect3D();

   // Here we unregister the window class with the OS.
   UnregisterClass("UGPClass", windowClass.hInstance);

   return (int)msg.wParam;
}


bool InitializeDirect3D(HWND hwnd, bool fullscreen)
{
   // This object will allow us to set the display mode of the screen.
   D3DDISPLAYMODE DisplayMode;

   // This will allow us to set the parameters of the screen.
   D3DPRESENT_PARAMETERS Present_Parameters;

   // This is used to get the capabilities of the hardware.
   D3DCAPS9 D3DCaps;

   // It is always a good idea to clear out memory in object although not necessary.
   ZeroMemory(&Present_Parameters, sizeof(Present_Parameters));

   // Create the Direct3D object to get everything started.
   Direct3D_Object = Direct3DCreate9(D3D_SDK_VERSION);

   // Error checking.  Make sure that it was successful.
   if(Direct3D_Object == NULL)
      {
         MessageBox(NULL, "Error, couldn't initialize DirectX!?!",
                    "Error!", MB_OK);
         return false;
      }

   // This function will get the display mode of the device and place it in DisplayMode.
   if(FAILED(Direct3D_Object->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &DisplayMode)))
      {
         MessageBox(NULL, "Error setting the display mode.", "Error!", MB_OK);
         return false;
      }

   // Get the capabilities of the hardware.
	if(FAILED(Direct3D_Object->GetDeviceCaps(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, &D3DCaps)))
		{
		   return false;
	   }

   // Test which is supported, hardware or software vertex processing.
	DWORD VertexProcessing = 0;

	if(D3DCaps.VertexProcessingCaps != 0)
		VertexProcessing |= D3DCREATE_HARDWARE_VERTEXPROCESSING;
	else
		VertexProcessing |= D3DCREATE_SOFTWARE_VERTEXPROCESSING;

   // Here we are setting the applications parameters...
   if(fullscreen)
      {
         Present_Parameters.Windowed = FALSE;               // Window mode (fullscreen).
         Present_Parameters.BackBufferWidth = 640;
         Present_Parameters.BackBufferHeight = 480;
      }
   else
      Present_Parameters.Windowed = TRUE;                   // Window mode (not fullscreen).
   Present_Parameters.SwapEffect = D3DSWAPEFFECT_DISCARD;   // Dealing with animation (see doc).
   Present_Parameters.BackBufferFormat = DisplayMode.Format;// Render to the area of the screen.
   Present_Parameters.BackBufferCount = 1;                  // Number of back buffers.
   Present_Parameters.EnableAutoDepthStencil = TRUE;        // Check documentation.
   Present_Parameters.AutoDepthStencilFormat = D3DFMT_D16;  // Check documentation.


   // Now we must create the rendering device.
   if(FAILED(Direct3D_Object->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hwnd,
                                           D3DCREATE_SOFTWARE_VERTEXPROCESSING,
                                           &Present_Parameters, &D3D_Device)))
      {
         MessageBox(NULL, "CreateDevice() failed!  Make sure you have DirectX 9.",
                    "Error!", MB_OK);
         return false;
      }


   // One last check to be sure.
   if(D3D_Device == NULL)
      {
         MessageBox(NULL, "D3D_Device is equal to NULL!?!", "Error!", MB_OK);
         return false;
      }


   // Turn off lighting and culling.
   D3D_Device->SetRenderState(D3DRS_LIGHTING, FALSE);
	D3D_Device->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
   D3D_Device->SetRenderState(D3DRS_ZENABLE, TRUE);


   // Here we set up the object this demo will render (for example a triangle).
   // This will load the data into the vertex buffer.
   if(!InitializeObject())
      return false;

   // Lets get the capabilities of the hardware.
   D3D_Device->GetDeviceCaps(&D3DCaps);

   // With that information lets check and make sure that the hardware supports
   // vertex shader 2.0.
   if(D3DCaps.VertexShaderVersion < D3DVS_VERSION(2, 0))
      return false;

   // This will declare the vertex shader declartion (how the shader is set up).
   if(D3D_Device->CreateVertexDeclaration(SDec, &ShaderDeclaration) != D3D_OK)
      return false;

   // This will hold the source of the vertex shader.
	ID3DXBuffer *source = NULL;

   // Load the vertex shader into the source object..
	if(D3DXAssembleShaderFromFile("VertexShader1.vsh", NULL, NULL, NULL, &source, NULL) != D3D_OK)
      return false;

   // Now we create the vertex shader for Direct3D to use.
   if(D3D_Device->CreateVertexShader((DWORD*)source->GetBufferPointer(),&VertexShader) != D3D_OK)
      return false;

   // We can Release this now because we are done.
   if(source != NULL)
    {
        source->Release(); 
        source = NULL; 
    }

   // Set the projection matrix.
   D3DXMatrixPerspectiveFovLH(&ProjectionMatrix, 45.0f, 640/480, 0.0f, 500.0f);
   D3D_Device->SetTransform(D3DTS_PROJECTION, &ProjectionMatrix);

   return true;
}


bool InitializeObject()
{
   // Triangle data for its position and color.  Each point has a x, y, and z position
   // as well as a Red, Green, and Blue color (D3DCOLOR_XRGB(Red, Green, Blue).
	Vertex Triangle_Data[3] =
   {
		{-0.5f, -0.5f, 0.0f, D3DCOLOR_XRGB(255,0,0)},
      {0.5f, -0.5f, 0.0f, D3DCOLOR_XRGB(255,0,0)},
	   {0.0f, 0.3f, 0.0f, D3DCOLOR_XRGB(0,0,255)}
	};

   // Create the vertex buffer that will hold the triangle.
	if(FAILED(D3D_Device->CreateVertexBuffer(sizeof(Triangle_Data),
                                            0, D3DFVF_D3DVertex,
                                            D3DPOOL_DEFAULT, &Vertex_Buffer, NULL)))
      {
         return false;  // Exit this function if there is some kind of problem.
      }
	
   // Pointer to the buffer.
   Vertex* Vertices;

   // Lock the buffer we can write to it.
   if(FAILED(Vertex_Buffer->Lock(0, sizeof(Triangle_Data), (void**)&Vertices, 0)))
      return false;

   // Here we copy the triangle's data into the vertex buffer.
   memcpy(Vertices, Triangle_Data, sizeof(Triangle_Data));

   // Unlock when your done coping data into the buffer.
   Vertex_Buffer->Unlock();

   return true;
}


void SetShader()
{
   D3DXMATRIX matTrans; // Translation.
	D3DXMATRIX matRot;   // Rotation.

   // What we will do is create a model view projection matrix.  This matrix
   // will be used by the shader to translate the scene.  In the matrices
   // demo we used a Direct3D function to translate but I decided to use
   // the shader so you could see how to set vertex shader values.  To
   // get the matrix we have to multiply the world, view, and projection
   // matrices together.  First we will do so then set the matrix in the shader.
   // Lastly we will activate the vertex shader for this frame.

   // Here we will set the translate we want on the scene.
   D3DXMatrixTranslation(&matTrans, 0.0f, 0.0f, 3.0f);
   // This will create a rotation matrix.
	D3DXMatrixRotationYawPitchRoll(&matRot, D3DXToRadian(0.0), D3DXToRadian(0.0), 0.0f);
   // Calculate the world matrix.  Which is the rotation * translation.
   WorldMatrix = matRot * matTrans;

   // Load the view matrix into ViewMatrix because we will also need that..
   D3DXMatrixIdentity(&ViewMatrix);

   // Set the model view proj matrix.  We already have the projection matrix from the
   // initialize so lets multiply all of them together.
   D3DXMATRIX modelViewProj = WorldMatrix * ViewMatrix * ProjectionMatrix;

   // Transpose the model view matrix and we are all set.
   D3DXMatrixTranspose(&modelViewProj, &modelViewProj);

   // Load the model view projection matrix in registers c[0]-c[3].  If you look in the
   // shader file you would see that it expects the model view proj to be in those registers.
   // SetVertexShaderConstantF expects the constant register index (c[0]), the variable
   // you want to send, and how many registers it will take (c[0] through c[3]).  In the
   // shader the matrix (c[0] - c[3]) is combined with the vertex's x, y, z that is defined
   // in the triangle data array to create its new position.
   D3D_Device->SetVertexShaderConstantF(0, (float*)modelViewProj, 4);

   // Set how the vertex shader is defined.
   D3D_Device->SetVertexDeclaration(ShaderDeclaration);

   // Set the vertex shader we loaded.
   D3D_Device->SetVertexShader(VertexShader);
}


void RenderScence()
{
   // Clear the screen.
   D3D_Device->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,
                     D3DCOLOR_XRGB(0,0,0), 1.0f, 0);
    
   // Tell Direct 3D we are now going to start drawing.
   D3D_Device->BeginScene();

	   // This will bind the vertex data in the buffer to the Direct3D device.
      D3D_Device->SetStreamSource(0, Vertex_Buffer, 0, sizeof(Vertex));

      // Set the vertex shader variables.  This must be called before DrawPrimitive();
      // Basically are giving the vertex shader all the info it needs to work.
      SetShader();

      // This will draw everything in the buffer.
      D3D_Device->DrawPrimitive(D3DPT_TRIANGLELIST, 0, 1);
    
   // We are done drawing for this scene.
   D3D_Device->EndScene();
    
   // Swap the old frame with the new one.
   D3D_Device->Present(NULL, NULL, NULL, NULL);
}


void ShutdownDirect3D()
{
   // Here we release the Direct3D objects and vertex shader.
   if(VertexShader != NULL)
      {
         VertexShader->Release();
         VertexShader = NULL;
      }

   if(ShaderDeclaration != NULL)
      {
         ShaderDeclaration->Release();
         ShaderDeclaration = NULL;
      }

   if(D3D_Device != NULL)
      {
         D3D_Device->Release();
         D3D_Device = NULL;
      }

   if(Direct3D_Object != NULL)
      {
         Direct3D_Object->Release();
         Direct3D_Object = NULL;
      }

   if(Vertex_Buffer != NULL)
      {
	      Vertex_Buffer->Release();
         Vertex_Buffer = NULL;
      }
}


// Recap:

// There you have it.  A simple vertex shader application in Direct3D.  If you
// are unfamilar with assembly language code then the first time you saw the shader
// code it most likely scared you a little (it scared me the first time) but over
// time the shader code gets much easier.  Basically you can do almost what you want
// in a vertex shader (dealing with a vertex).  Next demo we will create a simple
// diffuse lighting in a vertex shader.  To create a vertex shader is pretty simple as
// you see.  First you check for shader support then you will load the shader into
// Direct3D.  After that all you have to do is start of shader in the rendering function
// and make sure you loaded all the values the shader needs before you start the actual
// drawing.  Lastly when the app is done you just delete everything you used.  If you don't
// supply the vertex shader with the information it is expecting ahead of time you could
// cause a lot of error or unexpected results.  Its almost like the following...

// int A, B, C;
//
// C = A + B;

// Now if you was to print C there is no telling what would happen because the variables were
// never initialized.  Same thing with the shader constants.


// Copyright September 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com